import os
import unittest
from django import setup
from kliko.testutil import kliko_data

# we need to init django before we can import django_form
os.environ['DJANGO_SETTINGS_MODULE'] = 'test.test_settings'
setup()

import kliko.django_form


class TestDjangoForm(unittest.TestCase):
    def test_django_form(self):
        Form = kliko.django_form.generate_form(kliko_data)
        form = Form()
        form.is_valid()



import os
import unittest
import docker
import docker.utils
from builtins import open

import kliko.docker
from kliko.exceptions import KlikoException

TEST_IMAGE = 'radioastro/klikotest'


PARAMS_FILE = os.path.join(os.path.dirname(__file__), '../examples/kliko.yml')


class TestUtils(unittest.TestCase):
    def setUp(self):
        config = docker.utils.kwargs_from_env()
        config['version'] = "1.20"
        self.client = docker.Client(**config)
        self.assertTrue(self.client.images(name=TEST_IMAGE), "docker image %s not found" % TEST_IMAGE)

    def test_extract(self):
        image_params = kliko.docker.extract_params(self.client, TEST_IMAGE)
        true_params = open(PARAMS_FILE, mode='r', encoding='utf-8').read()
        self.assertEqual(image_params, true_params)

    def test_extract_without_params(self):
        with self.assertRaises(KlikoException):
            kliko.docker.extract_params(self.client, 'python:3.5')



from os import path
import yaml
import json

here = path.dirname(path.realpath(__file__))
kliko_file = path.join(here, "../examples/kliko.yml")
parameters_file = path.join(here, "../examples/parameters.json")

with open(kliko_file, 'r') as f:
    kliko_data = yaml.load(f)

with open(parameters_file, 'r') as f:
    parameters_data = json.load(f)

parameters_str = json.dumps(parameters_data)



"""
Command line utilities for Kliko
"""
from __future__ import absolute_import
import argparse
import json
import os
import sys
import tempfile
from shutil import copyfile

import docker
import docker.utils
import requests.exceptions
import yaml

from kliko.docker import extract_params
from kliko.validate import validate_kliko


def directory_exists(path):
    """check if a directory exists
    """
    if not os.path.isdir(path):
        raise argparse.ArgumentTypeError("directory doesn't exists: '%s'" % path)
    return path


def file_exists(path):
    """check if a file exists
    """
    if not os.path.isfile(path):
        raise argparse.ArgumentTypeError("file doesn't exists: '%s'" % path)
    return path


def generate_kliko_cli_parser(kliko_data, parent_parser=None):
    """
    Generate a command line parser from a Kliko structure.

    Args:
        kliko_data (dict): A nested kliko structure

    Returns:
        argparse.ArgumentParser: a configured argument parser

    """
    if parent_parser:
        parser = argparse.ArgumentParser(parents=[parent_parser])
    else:
        parser = argparse.ArgumentParser(description=kliko_data['description'])

    for section in kliko_data['sections']:
        for field in section['fields']:

            args = ['--' + field['name']]

            kwargs = {'dest': field['name']}

            type_ = field['type']
            if type_ in ('float', 'bool', 'int'):
                kwargs['type'] = eval(field['type'])
            elif type_ == 'char':
                kwargs['type'] = str
            elif type_ == 'file':
                kwargs['type'] = file_exists
            elif type_ == 'choice':
                kwargs['type'] = str
                kwargs['choices'] = field['choices']

            help = ""

            if 'label' in field:
                help += field['label']
                if 'help_text' in field:
                    help += ", "

            if 'help_text' in field:
                help += field['help_text']

            if 'initial' in field:
                kwargs['default'] = field['initial']
                help += " (default: %s)" % field['initial']

            if 'required' in field:
                kwargs['required'] = field['required']

            if help:
                kwargs['help'] = help

            parser.add_argument(*args, **kwargs)
    return parser


def prepare_io(parameters, target=None):
    """
    args:
        parameters: A dict containing the parameters
        target: output or work dir for kliko
    returns:
        tuple: (path to parameters file, path to target folder, path to input folder)
    """
    here = os.getcwd()

    if not target:
        target = os.path.join(here, 'output')

    if not os.path.exists(target):
        os.mkdir(target)

    if sys.platform == "darwin":
        # tempfolder not mounted into docker virtual machine
        parameters_path = os.path.join(here, 'parameters.json')
        parameters_file = open(parameters_path, 'w')
        input_ = os.path.join(here, 'input')
        if not os.path.exists(input_):
            os.mkdir(input_)
    else:
        _, parameters_path = tempfile.mkstemp()
        parameters_file = open(parameters_path, 'w')
        input_ = tempfile.mkdtemp()

    parameters_file.write(parameters)
    parameters_file.close()
    return parameters_path, target, input_


def first_parser(argv):
    """
    This is only used when script in invoked with 0 or 1 args (should be kliko image name).
    """
    epilogue = """
This is the command line based Kliko Image runner. It enables you to run a container directly from the console.

Use:

  $ kliko-run <kliko-image> --help

to see the list of accepted arguments for the kliko image. But default kliko-run will create an output folder
in the current working folder where the results of the Kliko run are written to.

Note that you still need to download (docker pull) or build the Kliko image yourself.
"""

    parser = argparse.ArgumentParser(description='Kliko image runner', epilog=epilogue)
    parser.add_argument('image_name', type=str, help='Kliko image you want to run')
    help = 'Kliko parameters. Run --help after image_name to retreive list from image'
    parser.add_argument('IMAGE OPTIONS', type=str, nargs='*', help=help)

    parsed = parser.parse_args(argv[1:2])
    return parsed.image_name


def second_parser(argv, kliko_data):
    """
    Used when kliko image is known, so we can extract the parameters.
    """
    # we recreate parser since otherwise we have a double help conflict
    no_help_parser = argparse.ArgumentParser(description='kliko runner', add_help=False)
    no_help_parser.add_argument('image_name', type=str)
    no_help_parser.add_argument('--target_folder', type=str)

    final_parser = generate_kliko_cli_parser(kliko_data, no_help_parser)
    final_parsed = final_parser.parse_args(argv[1:])
    target = final_parsed.target_folder
    parameters = vars(final_parsed)
    parameters.pop('image_name')
    parameters.pop('target_folder')
    return parameters, target


def kliko_runner(argv):
    image_name = first_parser(argv)

    config = docker.utils.kwargs_from_env()
    if 'tls' in config:
        config['tls'].verify = False   # TODO: disalbe this for now
    docker_client = docker.Client(**config)

    try:
        raw_kliko_data = extract_params(docker_client, image_name)
    except requests.exceptions.ConnectionError:
        print("Can't connect to docker daemon (config '%s')" % str(config))
        exit(1)
    except docker.errors.NotFound:
        print("Can't find docker image '%s'" % image_name)
        exit(1)

    kliko_data = validate_kliko(yaml.safe_load(raw_kliko_data))

    parameters, target = second_parser(argv, kliko_data)
    parameters_string = json.dumps(parameters)
    parameters_file, target, input_ = prepare_io(parameters_string, target=target)

    files = []
    for section in kliko_data['sections']:
        for field in section['fields']:
            if field['type'] == 'file':
                files.append((field['name'], parameters[field['name']]))

    if kliko_data['io'] == 'split':
        binds = [
            input_ + ':/input:ro',
            target + ':/output:rw',
            parameters_file + ':/parameters.json:ro',
        ]
        for fieldname, path in files:
            copyfile(path, os.path.join(input_, fieldname))
    else:
        binds = [
            target + ':/work:rw',
            parameters_file + ':/parameters.json:ro',
        ]
        for fieldname, path in files:
            copyfile(path, os.path.join(target, fieldname))

    host_config = docker_client.create_host_config(binds=binds)

    container = docker_client.create_container(image=image_name, host_config=host_config)
    docker_client.start(container)
    error_code = docker_client.wait(container)
    warnings = container.get('Warnings')
    if warnings:
        for warning in warnings:
            print(warning)
    stdout = docker_client.logs(container, stdout=True, stream=False).decode('utf-8')
    docker_client.remove_container(container)  # always clean up the container
    print(stdout)
    if error_code != 0:
        exit(1)



import unittest
from pykwalify.errors import SchemaError
import kliko.validate
from kliko.testutil import kliko_data, parameters_data, kliko_file, parameters_file


class TestExample(unittest.TestCase):

    def test_validate_kliko(self):
        kliko.validate.validate_kliko(kliko_data)

    def test_convert_to_parameters_schema(self):
        kliko.validate.convert_to_parameters_schema(kliko_data)

    def test_validate_parameters(self):
        kliko.validate.validate_parameters(parameters_data, kliko_data)

    def test_validate(self):
        kliko.validate.validate(kliko_file, parameters_file)

    def test_validate_kliko_file_missing(self):
        with self.assertRaises(IOError):
            kliko.validate.validate('/non_existing_kliko_file', parameters_file)

    def test_validate_parameters_file_missing(self):
        with self.assertRaises(IOError):
            kliko.validate.validate(kliko_file, '/non_existing_parameters_file')

    def test_missing_field(self):

        kliko_data = {'sections': [{'name': 'section',
                                    'fields': [{'name': 'required',
                                                'type': 'float',
                                                'required': True}]
                                    }]
                      }
        parameters_data = {}
        with self.assertRaises(SchemaError):
            kliko.validate.validate_parameters(parameters_data, kliko_data)

    def test_illegal_field(self):
        kliko_data = {'sections': [{'name': 'section',
                                    'fields': [{'name': 'field',
                                                'type': 'float'}]
                                    }]
                      }
        parameters_data = {'illegal_field': 'bla'}
        with self.assertRaises(SchemaError):
            kliko.validate.validate_parameters(parameters_data, kliko_data)



"""
Kliko and parameter validation related functions.
"""
from pykwalify.core import Core
import yaml
import json
import os


here = os.path.dirname(os.path.realpath(__file__))

SCHEMA_VERSION = 2


def validate_kliko(kliko, version=SCHEMA_VERSION):
    """
    validate a kliko yaml string

    args:
        kliko: a parsed kliko object

    returns:
        dict: a (nested) kliko structure

    raises:
        an exception if the string can't be parsed or is not in the following the Kliko schema
    """
    # first try to parse it, to make sure it is parsable

    schema_file = os.path.join(here, "schemas/%s.yml" % version)
    c = Core(source_data=kliko, schema_files=[schema_file])
    c.validate(raise_exception=True)
    return kliko


def convert_to_parameters_schema(kliko):
    """
    Convert a kliko schema into a validator for the parameters generated with a kliko schema.

    args:
        kliko (str): a kliko definition
    returns:
        A structure for a pykwalify validator
    """

    type_map = {
        'choice': 'str',
        'char': 'str',
        'file': 'str',
    }

    mapping = {}

    for section in kliko['sections']:
        for field in section['fields']:
            # TODO: we can't define multiple types:
            # https://github.com/Grokzen/pykwalify/issues/39
            # type_ = type_map.get(field['type'], field['type'])
            value = {'type': 'any', 'required': 'required' in field}
            mapping[field['name']] = value

    return {'type': 'map', 'mapping': mapping}


def validate_parameters(parameters, kliko):
    """
    validate a set of parameters given a kliko definition

    args:
        parameters (dict): A structure that should follow the given kliko structure
        kliko (dict): A nested dict which defines the valid parameters in Kliko format

    returns:
        str: the parsed parameters

    raises:
        an exception if the string can't be parsed or is not in the defining valid parameters
    """
    schema = convert_to_parameters_schema(kliko)
    c = Core(source_data=parameters, schema_data=schema)
    c.validate(raise_exception=True)
    return True


def validate(kliko_file='/kliko.yml', paramaters_file='/parameters.json'):
    """
    Validate the kliko and paramaters file and parse the parameters file. Should be run inside
    the Kliko container.

    Args:
        kliko_file (str):  Path to a kliko file
        paramaters_file (str): path to a parameters file

    Returns:
        The validated and parsed paramaters file
    """

    with open(kliko_file, 'r') as f:
        kliko = yaml.safe_load(f)

    validate_kliko(kliko)

    with open(paramaters_file, 'r') as f:
        parameters = json.load(f)

    validate_parameters(parameters, kliko)

    return parameters



import unittest
import os
import kliko.cli
from kliko.testutil import kliko_data, parameters_str

this_file = os.path.realpath(__file__)

class TestCli(unittest.TestCase):
    def test_cli(self):
        parser = kliko.cli.generate_kliko_cli_parser(kliko_data)

    def test_prepare_io(self):
        kliko.cli.prepare_io(parameters_str)

    def test_first_parser(self):
        with self.assertRaises(SystemExit):
            kliko.cli.first_parser(['kliko-run'])
        with self.assertRaises(SystemExit):
            kliko.cli.first_parser(['kliko-run', '--help'])
        kliko.cli.first_parser(['kliko-run', 'radioastro/klikotest'])

        kliko.cli.first_parser(['kliko-run', 'radioastro/klikotest', '--help'])

    def test_second_parser(self):
        with self.assertRaises(SystemExit):
            kliko.cli.second_parser(['kliko-run'], kliko_data)
        with self.assertRaises(SystemExit):
            kliko.cli.second_parser(['kliko-run', '--help'], kliko_data)
        with self.assertRaises(SystemExit):
            kliko.cli.second_parser(['kliko-run', 'radioastro/klikotest'], kliko_data)
        with self.assertRaises(SystemExit):
            kliko.cli.second_parser(['kliko-run', 'radioastro/klikotest', '--help'], kliko_data)

        kliko.cli.second_parser(['kliko-run', 'radioastro/klikotest', '--choice', 'second', '--char', 'gijs',
                                 '--file', this_file, '--int', '10'], kliko_data)

    def test_kliko_runner(self):
        with self.assertRaises(SystemExit):
            kliko.cli.kliko_runner(['kliko-run'])
        with self.assertRaises(SystemExit):
            kliko.cli.kliko_runner(['kliko-run', '--help'])
        with self.assertRaises(SystemExit):
            kliko.cli.kliko_runner(['kliko-run', 'radioastro/klikotest'])
        with self.assertRaises(SystemExit):
            kliko.cli.kliko_runner(['kliko-run', 'radioastro/klikotest', '--help'])

        kliko.cli.kliko_runner(['kliko-run', 'radioastro/klikotest', '--choice', 'second', '--char', 'gijs',
                                 '--file', this_file, '--int', '10'])


from setuptools import setup, find_packages

__version__ = "0.6.1"


install_requires = [
    'pykwalify',
    'PyYAML',
    'future',
    'docopt'
]

extras_require = {
    'django': ['django', 'django-form-utils'],
    'docker': ['docker-py'],
}

scripts = [
    'bin/kliko-validate',
    'bin/kliko-run',
]

setup(
    name="kliko",
    version=__version__,
    packages=find_packages(),
    scripts=scripts,
    install_requires=install_requires,
    extras_require=extras_require,
    package_data={
        '': ['*.txt', '*.rst'],
        'kliko': ['schemas/*.yml'],
    },
    author="Gijs Molenaar",
    author_email="gijs@pythonic.nl",
    description="Scientific Compute Container Spec Parser",
    license="GPL2",
    keywords="science docker yaml json",
    url="https://github.com/gijzelaerr/kliko",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development :: Build Tools",
        "Topic :: System :: Distributed Computing",
        "Topic :: System :: Emulators",
        "Topic :: System :: Operating System Kernels :: Linux",
        ]
)
